import megamu.mesh.*;

class World {
  Voronoi voronoi;
  MPolygon[] regions;

  Delaunay delaunay;
  float[][] delaunayEdges;
  int[][] delaunayLinks;

  private ArrayList<Node> nodes;
  private ArrayList<Node> openNodes;
  Node startNode, goalNode;

  World(int numOfNode) {
    nodes = new ArrayList<Node>();
    openNodes = new ArrayList<Node>();
    for (int i = 0; i < numOfNode; i++) {
      nodes.add(new Node(new PVector(random(0, width), random(0, height))));
    }
    setStartNode(0);
    setGoalNode(nodeSize()-1);
  }

  void update() {
    resetNodes();

    if (goalNode.index > nodes.size() - 1) {
      setGoalNode(nodes.size() - 1);
    }

    for (int i = 0; i < nodeSize (); i++) {
      getNode(i).index = i;
    }

    //convert node list to float array
    float[][] points = new float[nodes.size()][2];
    for (int i = 0; i < nodes.size (); i++) {
      points[i][0] = nodes.get(i).position.x;
      points[i][1] = nodes.get(i).position.y;
    }

    voronoi = new Voronoi(points);
    regions = voronoi.getRegions();

    delaunay = new Delaunay(points);
    delaunayEdges = delaunay.getEdges();
    delaunayLinks = delaunay.getLinks();

    solveStart();

    //detect route
    for (int i=0; i < nodeSize (); i++) {
      Node n = getNode(i);
      if (n.nodeType == NodeType.GOAL && n.done) {
        n.isRoute = true;
        Node p = n.prev;
        while (p != null) {
          p.isRoute = true;
          p = p.prev;
        }
      }
    }
  }

  public Node getNode(int index) {
    return nodes.get(index);
  }

  public void removeNode(int index) {
    if (nodes.size() > 2) {
      nodes.remove(index);
    }
  }

  public void addNode(Node node) {
    //if already exist at same position
    for(int i=0;i<nodeSize();i++){
      if(node.position.x == getNode(i).position.x && node.position.y == getNode(i).position.y){
        return;
      }
    }
    nodes.add(node);
  }

  public int nodeSize() {
    return nodes.size();
  }

  private void resetNodes() {
    openNodes.clear();
    for (int i = 0; i < nodeSize (); i++) {
      getNode(i).reset();
    }
  }

  private float heuristicAt(PVector position) {
    return position.dist(goalNode.position);
  }

  private void solveStart() {
    startNode.score = 0;
    startNode.fs = startNode.score + heuristicAt(startNode.position);
    openNodes.add(startNode);
    solveNext();
  }

  void solveNext() {
    float minScore = Float.MAX_VALUE;
    int minIndex = -1;
    Node u = null;
    for (int i = 0; i < openNodes.size (); i++) {
      Node targetNode = openNodes.get(i);
      if (targetNode.done) continue;
      if (targetNode.fs < minScore) {
        minScore = targetNode.fs;
        minIndex = i;
        u = targetNode;
      }
    }

    if (u == null) {
      return;
    }

    openNodes.remove(minIndex);
    u.done = true;

    if (u.nodeType == NodeType.GOAL) {
      return;
    }

    int[] linkNodes = delaunay.getLinked(u.index);
    for (int i = 0; i < linkNodes.length; i++) {
      Node v = getNode(linkNodes[i]);
      if (v.done || v.nodeType == NodeType.OBSTACLE) continue;
      //following number 1 is moving cost between node to node.
      float cost = dist(u.position.x, u.position.y, v.position.x, v.position.y);
      if (u.score + cost + heuristicAt(v.position) < v.fs) {
        v.score = u.score + cost;
        v.fs = v.score + heuristicAt(v.position);
        v.prev = u;

        if (openNodes.indexOf(v) == -1)openNodes.add(v);
      }
    }
    solveNext();
  }

  public void setStartNode(int index) {
    if (getNode(index).nodeType == NodeType.START || getNode(index).nodeType == NodeType.GOAL) return;
    if (startNode != null) {
      startNode.nodeType = NodeType.NONE;
    }
    getNode(index).nodeType = NodeType.START;
    startNode = getNode(index);
  }

  public void setGoalNode(int index) {
    if (getNode(index).nodeType == NodeType.START || getNode(index).nodeType == NodeType.GOAL) return;
    if (goalNode != null) {
      goalNode.nodeType = NodeType.NONE;
    }
    getNode(index).nodeType = NodeType.GOAL;
    goalNode = getNode(index);
  }

  public void setObstacleNode(int index) {
    Node n = getNode(index);
    if (n.nodeType != NodeType.START && n.nodeType != NodeType.GOAL) {
      if (n.nodeType == NodeType.NONE) {
        n.nodeType = NodeType.OBSTACLE;
      } else if (n.nodeType == NodeType.OBSTACLE) {
        n.nodeType = NodeType.NONE;
      }
    }
  }
}

